"""Define a request/response connection to an AV Receiver."""
import logging
from abc import ABC
from collections import defaultdict

logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(__name__)


class HTTPConnection(ABC):
    """Define the HTTP connection interface."""

    def __init__(self, host: str, upnp_data=None):
        """Init the connection."""
        self.host = host
        self._upnp_data = upnp_data
        self.port = None  # type: int
        self._device_info_url = None  # type: str
        self._device_info = defaultdict(None)

    @property
    def device_info(self):
        """Return the device info dict."""
        return self._device_info



"""Defines the dispatch component for notifying others of signals."""
import asyncio
import functools
from collections import defaultdict
from typing import Any, Callable, Dict, List, Sequence

TargetType = Callable[..., Any]
DisconnectType = Callable[[], None]
ConnectType = Callable[[str, TargetType], DisconnectType]
SendType = Callable[..., Sequence[asyncio.Future]]


class Dispatcher:
    """Define the dispatch class."""

    def __init__(
        self,
        *,
        connect: ConnectType = None,
        send: SendType = None,
        signal_prefix: str = "",
        loop=None
    ):
        """Init a new dispatch component."""
        self._connect = connect or self._default_connect
        self._send = send or self._default_send
        self._signal_prefix = signal_prefix
        self._loop = loop or asyncio.get_event_loop()
        self._signals = defaultdict(list)
        self._disconnects = []

    def connect(self, signal: str, target: TargetType) -> DisconnectType:
        """Connect function to signal.  Must be ran in the event loop."""
        disconnect = self._connect(self._signal_prefix + signal, target)
        self._disconnects.append(disconnect)
        return disconnect

    def send(self, signal: str, *args: Any) -> Sequence[asyncio.Future]:
        """Fire a signal.  Must be ran in the event loop."""
        return self._send(self._signal_prefix + signal, *args)

    def disconnect_all(self):
        """Disconnect all connected."""
        disconnects = self._disconnects.copy()
        self._disconnects.clear()
        for disconnect in disconnects:
            disconnect()

    def _default_connect(self, signal: str, target: TargetType) -> DisconnectType:
        """Connect function to signal.  Must be ran in the event loop."""
        self._signals[signal].append(target)

        def remove_dispatcher() -> None:
            """Remove signal listener."""
            try:
                self._signals[signal].remove(target)
            except ValueError:
                # signal was already removed
                pass

        return remove_dispatcher

    def _default_send(self, signal: str, *args: Any) -> Sequence[asyncio.Future]:
        """Fire a signal.  Must be ran in the event loop."""
        targets = self._signals[signal]
        futures = []
        for target in targets:
            task = self._call_target(target, *args)
            futures.append(task)
        return futures

    def _call_target(self, target, *args) -> asyncio.Future:
        check_target = target
        while isinstance(check_target, functools.partial):
            check_target = check_target.func
        if asyncio.iscoroutinefunction(check_target):
            return self._loop.create_task(target(*args))
        return self._loop.run_in_executor(None, target, *args)

    @property
    def signals(self) -> Dict[str, List[TargetType]]:
        """Get the dictionary of registered signals and callbacks."""
        return self._signals



"""Define the Denon/Marantz telnet connection."""
import logging
from datetime import datetime
from typing import Optional

import aiofiles
import yaml

from pyavreceiver.denon import const as denon_const
from pyavreceiver.denon.commands import get_command_lookup
from pyavreceiver.denon.response import DenonMessage
from pyavreceiver.telnet_connection import TelnetConnection

_LOGGER = logging.getLogger(__name__)


class DenonTelnetConnection(TelnetConnection):
    """Define the simplest Denon telnet connection."""

    def __init__(
        self,
        avr,
        host,
        *,
        port: int = denon_const.CLI_PORT,
        timeout: float = denon_const.DEFAULT_TIMEOUT,
        heart_beat: Optional[float] = denon_const.DEFAULT_HEART_BEAT,
    ):
        """Init the connection."""
        super().__init__(avr, host, port=port, timeout=timeout, heart_beat=heart_beat)
        self._message_interval_limit = denon_const.MESSAGE_INTERVAL_LIMIT

    async def _load_command_dict(self, path=None):
        async with aiofiles.open("pyavreceiver/denon/commands.yaml") as file:
            self._command_dict = yaml.safe_load(await file.read())

    def _get_command_lookup(self, command_dict):
        return get_command_lookup(command_dict)

    async def _response_handler(self):
        while True:
            msg = None  # temporary for error detection
            try:
                msg = await self._reader.readuntil(
                    separator=denon_const.TELNET_SEPARATOR.encode()
                )
                message = msg.decode()[:-1]
                self._last_activity = datetime.utcnow()
                resp = DenonMessage(message, command_dict=self._command_dict)
                self._handle_event(resp)

                # Check if this is a response to a previous command
                if commands := self._expected_responses.get(resp.command):
                    commands.popleft()
                    if not commands:
                        del commands
            # pylint: disable=broad-except, fixme
            except Exception as err:
                # TODO: error handling
                _LOGGER.critical(err)
                _LOGGER.critical(msg)
                raise err



"""Define an audio/video receiver."""
from collections import defaultdict
from typing import Optional

from pyavreceiver import const
from pyavreceiver.command import CommandValues
from pyavreceiver.dispatch import Dispatcher
from pyavreceiver.telnet_connection import TelnetConnection
from pyavreceiver.zone import Zone


class AVReceiver:
    """Representation of an audio/video receiver."""

    def __init__(
        self,
        host: str,
        *,
        telnet: bool = True,
        http: bool = True,
        http_api=None,
        upnp: bool = True,
        timeout: float = const.DEFAULT_TIMEOUT,
        heart_beat: Optional[float] = const.DEFAULT_HEART_BEAT,
        dispatcher: Dispatcher = Dispatcher(),
        main_zone: Zone = Zone,
        aux_zone: Zone = Zone,
    ):
        """Init the device."""
        self._host = host
        self._connection = None  # type: TelnetConnection
        self._dispatcher = dispatcher
        self._telnet = telnet
        self._http = http
        self._http_connection = http_api
        self._upnp = upnp
        self._connections = []
        self._state = defaultdict()
        self._sources = None  # type: dict
        self._main_zone = None  # type: Zone
        self._main_zone_class = main_zone
        self._aux_zone_class = aux_zone
        self._model_name = None
        self._mac_address = None
        self._zones = None

    async def init(
        self,
        *,
        auto_reconnect=False,
        reconnect_delay: float = const.DEFAULT_RECONNECT_DELAY,
    ):
        """Await the initialization of the device."""
        if self._http and self._http_connection:
            await self.update_device_info()
        await self._connection.init(
            auto_reconnect=auto_reconnect, reconnect_delay=reconnect_delay
        )
        self._main_zone = self._main_zone_class(self)
        # pylint: disable=protected-access
        if self._sources:
            self._connection._command_lookup[const.ATTR_SOURCE]._values = CommandValues(
                self._sources
            )

    async def connect(
        self,
        *,
        auto_reconnect=False,
        reconnect_delay: float = const.DEFAULT_RECONNECT_DELAY,
    ):
        """Connect to the audio/video receiver."""
        if self._telnet:
            await self._connection.connect_telnet(
                auto_reconnect=auto_reconnect, reconnect_delay=reconnect_delay
            )
            self._connections.append(self._connection.disconnect_telnet)

    async def disconnect(self):
        """Disconnect from the audio/video receiver."""
        while self._connections:
            disconnect = self._connections.pop()
            await disconnect()

    def update_state(self, state_update: dict) -> bool:
        """Handle a state update."""
        update = False
        for attr, val in state_update.items():
            if attr not in self._state or self._state[attr] != val:
                self._state[attr] = val
                update = True
        return update

    async def update_device_info(self):
        """Update information about the A/V Receiver."""
        info = await self._http_connection.get_device_info()
        self._model_name = info.get(const.INFO_MODEL)
        self._mac_address = info.get(const.INFO_MAC)
        self._zones = info.get(const.INFO_ZONES)

        sources = await self._http_connection.get_source_names()
        self._sources = sources

    @property
    def dispatcher(self) -> Dispatcher:
        """Get the dispatcher instance."""
        return self._dispatcher

    @property
    def connection_state(self) -> str:
        """Get the state of the connection."""
        return self._connection.state

    @property
    def host(self) -> str:
        """Get the host."""
        return self._host

    @property
    def main(self) -> Zone:
        """Get the main zone object."""
        return self._main_zone

    @property
    def state(self) -> defaultdict:
        """Get the current state."""
        return self._state

    @property
    def power(self) -> str:
        """The state of power."""
        return self.state.get(const.ATTR_POWER)

    def set_power(self, val: bool) -> bool:
        """Request the receiver set power to val."""
        # pylint: disable=protected-access
        command = self._connection._command_lookup[const.ATTR_POWER].set_val(val)
        self._connection.send_command(command)
        return True



"""Define a Denon/Marantz Audio Video Receiver."""
from typing import Optional

from pyavreceiver.denon import const as denon_const
from pyavreceiver.denon.telnet_connection import DenonTelnetConnection
from pyavreceiver.denon.zone import DenonMainZone
from pyavreceiver.dispatch import Dispatcher
from pyavreceiver.receiver import AVReceiver
from pyavreceiver.zone import Zone


class DenonReceiver(AVReceiver):
    """Representation of a Denon or Marantz A/V Receiver."""

    def __init__(
        self,
        host,
        *,
        telnet: bool = True,
        http: bool = True,
        http_api=None,
        upnp: bool = True,
        timeout: float = denon_const.DEFAULT_TIMEOUT,
        heart_beat: Optional[float] = denon_const.DEFAULT_HEART_BEAT,
        dispatcher: Dispatcher = Dispatcher(),
        main_zone: DenonMainZone = DenonMainZone,
        aux_zone: Zone = Zone
    ):
        super().__init__(
            host,
            telnet=telnet,
            http=http,
            http_api=http_api,
            upnp=upnp,
            timeout=timeout,
            heart_beat=heart_beat,
            dispatcher=dispatcher,
            main_zone=main_zone,
            aux_zone=aux_zone,
        )
        self._connection = DenonTelnetConnection(
            self,
            host,
            timeout=timeout,
            heart_beat=heart_beat,
        )



"""Define Denon/Marantz A/V Receiver Zones."""
from pyavreceiver.denon import const as denon_const
from pyavreceiver.zone import MainZone


class DenonMainZone(MainZone):
    """Implement a Denon A/V Receiver zone."""

    @property
    def audyssey_dynamic_eq(self) -> str:
        """The state of Audyssey dynamic EQ."""
        return self.get(denon_const.ATTR_DYNAMIC_EQ)

    def set_audyssey_dynamic_eq(self, val: bool) -> bool:
        """Request the receiver set Audyssey dynamic EQ to val."""
        return self.set(denon_const.ATTR_DYNAMIC_EQ, val)



"""Define the telnet message public interface."""
from abc import ABC, abstractmethod


class Message(ABC):
    """Define the telnet Message interface."""

    @property
    @abstractmethod
    def message(self) -> str:
        """Return the message."""

    @property
    @abstractmethod
    def raw_value(self) -> str:
        """Return the raw value from the message."""

    @property
    @abstractmethod
    def state_update(self) -> dict:
        """Return the derived state update."""

    @property
    @abstractmethod
    def command(self) -> str:
        """Return the command and parameter category."""

    @property
    def new_command(self) -> dict:
        """Return new command if it was parsed, else None."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the name. Maybe the same as command."""



"""Define persistent connection to an AV Receiver."""
import asyncio
import logging
from abc import ABC, abstractmethod
from collections import OrderedDict, defaultdict, deque
from datetime import datetime, timedelta
from typing import Optional

import telnetlib3

from pyavreceiver import const
from pyavreceiver.response import Message

logging.basicConfig(level=logging.DEBUG)
_LOGGER = logging.getLogger(__name__)


class TelnetConnection(ABC):
    """Define the telnet connection interface."""

    def __init__(
        self,
        avr,
        host: str,
        *,
        port: int = const.CLI_PORT,
        timeout: float = const.DEFAULT_TIMEOUT,
        heart_beat: Optional[float] = const.DEFAULT_HEART_BEAT,
    ):
        """Init the connection."""
        self._avr = avr
        self.host = host
        self.port = port
        self.commands = None
        self._command_dict = {}
        self._command_lookup = {}
        self._learned_commands = {}
        self.timeout = timeout  # type: int
        self._reader = None  # type: telnetlib3.TelnetReader
        self._writer = None  # type: telnetlib3.TelnetWriter
        self._response_handler_task = None  # type: asyncio.Task
        self._queued_commands = OrderedDict()
        self._command_queue_task = None  # type: asyncio.Task
        self._expected_responses = defaultdict(deque)
        self._sequence = 0  # type: int
        self._state = const.STATE_DISCONNECTED  # type: str
        self._auto_reconnect = True  # type: bool
        self._reconnect_delay = const.DEFAULT_RECONNECT_DELAY  # type: float
        self._reconnect_task = None  # type: asyncio.Task
        self._last_activity = datetime(1970, 1, 1)  # type: datetime
        self._last_command_time = datetime(2020, 1, 1)  # type: datetime
        self._heart_beat_interval = heart_beat  # type: Optional[float]
        self._heart_beat_task = None  # type: asyncio.Task
        self._message_interval_limit = const.MESSAGE_INTERVAL_LIMIT

    @abstractmethod
    async def _load_command_dict(self, path=None):
        """Load the commands YAML."""

    @abstractmethod
    def _get_command_lookup(self, command_dict):
        """Create a command lookup dict."""

    @abstractmethod
    async def _response_handler(self):
        """Handle messages received from the device."""

    def _heartbeat_command(self):
        command = self._command_lookup[const.ATTR_POWER].set_query()
        self.send_command(command, heartbeat=True)

    async def init(self, *, auto_reconnect: bool = True, reconnect_delay: float = -1):
        """Await the async initialization."""
        await self._load_command_dict()
        await self.connect(
            auto_reconnect=auto_reconnect, reconnect_delay=reconnect_delay
        )
        self._command_lookup = self._get_command_lookup(self._command_dict)

    async def connect(
        self, *, auto_reconnect: bool = False, reconnect_delay: float = -1
    ):
        """Connect to the AV Receiver - called by init only."""
        if self._state == const.STATE_CONNECTED:
            return
        if reconnect_delay < 0:
            reconnect_delay = self._reconnect_delay
        self._auto_reconnect = False
        await self._connect()
        self._auto_reconnect = auto_reconnect

    async def _connect(self):
        """Make Telnet connection."""
        try:
            open_future = telnetlib3.open_connection(self.host, self.port)
            self._reader, self._writer = await asyncio.wait_for(
                open_future, self.timeout
            )
        except Exception as error:
            raise error from error
        self._response_handler_task = asyncio.create_task(self._response_handler())
        self._state = const.STATE_CONNECTED
        self._command_queue_task = asyncio.create_task(self._process_command_queue())
        if self._heart_beat_interval is not None and self._heart_beat_interval > 0:
            self._heart_beat_task = asyncio.create_task(self._heart_beat())

        _LOGGER.debug("Connected to %s", self.host)
        self._avr.dispatcher.send(const.SIGNAL_TELNET_EVENT, const.EVENT_CONNECTED)

    async def disconnect(self):
        """Disconnect from the AV Receiver."""
        if self._state == const.STATE_DISCONNECTED:
            return
        if self._reconnect_task:
            self._reconnect_task.cancel()
            await self._reconnect_task
            self._reconnect_task = None
        await self._disconnect()
        self._state = const.STATE_DISCONNECTED

        _LOGGER.debug("Disconnected from %s", self.host)
        self._avr.dispatcher.send(const.SIGNAL_TELNET_EVENT, const.EVENT_DISCONNECTED)

    async def _disconnect(self):
        """Cancel response handler and pending tasks."""
        if self._heart_beat_task:
            self._heart_beat_task.cancel()
            try:
                await self._heart_beat_task
            except asyncio.CancelledError:
                pass
            self._heart_beat_task = None
        if self._response_handler_task:
            self._response_handler_task.cancel()
            try:
                await self._response_handler_task
            except asyncio.CancelledError:
                pass
            self._response_handler_task = None
        if self._command_queue_task:
            self._command_queue_task.cancel()
            try:
                await self._command_queue_task
            except asyncio.CancelledError:
                pass
            self._command_queue_task = None
        if self._writer:
            self._writer.close()
            self._writer = None
        self._reader = None
        self._sequence = 0
        self._queued_commands.clear()
        self._expected_responses.clear()

    async def _handle_connection_error(self, error: Exception = "hearbeat"):
        """Handle connection failures and schedule reconnect."""
        if self._reconnect_task:
            return
        await self._disconnect()
        if self._auto_reconnect:
            self._state = const.STATE_RECONNECTING
            self._reconnect_task = asyncio.create_task(self._reconnect())
        else:
            self._state = const.STATE_DISCONNECTED

        _LOGGER.debug("Disconnected from %s: %s", self.host, error)
        self._avr.dispatcher.send(const.SIGNAL_TELNET_EVENT, const.EVENT_DISCONNECTED)

    async def _reconnect(self):
        """Perform core reconnection logic."""
        # pylint: disable=broad-except
        while self._state != const.STATE_CONNECTED:
            try:
                await self._connect()
                self._reconnect_task = None
                return
            except Exception as err:
                # Occurs when we could not reconnect
                _LOGGER.debug("Failed to reconnect to %s: %s", self.host, err)
                await self._disconnect()
                await asyncio.sleep(self._reconnect_delay)
            except asyncio.CancelledError:
                # Occurs when reconnect is cancelled via disconnect
                return

    async def _heart_beat(self):
        """Check for activity and send a heartbeat to check for connection."""
        while self._state == const.STATE_CONNECTED:
            old_last_activity = self._last_activity
            last_activity = datetime.utcnow() - self._last_activity
            threshold = timedelta(seconds=self._heart_beat_interval)
            if last_activity > threshold:
                self._heartbeat_command()
                await asyncio.sleep(5)
                if self._last_activity <= old_last_activity:
                    await self._handle_connection_error()
            await asyncio.sleep(self._heart_beat_interval / 2)

    def send_command(self, command, heartbeat=False):
        """Execute a command."""
        if not heartbeat and self._state != const.STATE_CONNECTED:
            _LOGGER.debug(
                "Command failed %s - Not connected to device %s",
                command.message,
                self.host,
            )
            return

        _LOGGER.debug("queueing command: %s", command.message)

        self._queued_commands[command.command] = {
            "message": command.message,
            "command": command.command,
            "val": command.val,
            "attempt": 0,
        }

    async def _process_command_queue(self):
        while True:
            try:
                time_since_last_command = datetime.utcnow() - self._last_command_time
                threshold = timedelta(milliseconds=self._message_interval_limit)
                wait_time = self._message_interval_limit / 1000  # ms -> s
                if time_since_last_command > threshold:
                    next_command = None
                    try:
                        _, next_command = self._queued_commands.popitem(last=False)
                        if next_command is not None:
                            _LOGGER.debug("Popped command: %s", next_command["message"])
                            self._writer.write(next_command["message"])
                            await self._writer.drain()
                            self._last_command_time = datetime.utcnow()
                            wait_time = self._message_interval_limit / 1000 + 0.002
                            self._expected_responses[next_command["command"]].append(
                                ExpectedResponse(
                                    next_command["command"], next_command["val"]
                                )
                            )
                    except KeyError:
                        pass
                else:
                    wait_time = (
                        threshold.total_seconds()
                        - time_since_last_command.total_seconds()
                        + 0.002
                    )
                await asyncio.sleep(wait_time)
            # pylint: disable=broad-except, fixme
            except Exception as err:
                # TODO: error handling
                _LOGGER.critical(err)
                await asyncio.sleep(0.05)

    def _handle_event(self, resp: Message):
        """Handle a response event."""
        if resp.state_update == {}:
            _LOGGER.debug("No state update in message: %s", resp.message)
            return
        updated = self._avr.update_state(resp.state_update)
        expected_response_deque = self._expected_responses.get(resp.command)
        if expected_response_deque:
            try:
                expected_response = expected_response_deque.popleft()
                expected_response.set(resp)
            except IndexError:
                _LOGGER.debug("No expected response for: %s", resp.command)
        else:
            _LOGGER.debug("No expected response matched: %s", resp.command)
        if updated:
            self._avr.dispatcher.send(const.SIGNAL_STATE_UPDATE, resp.message)
        _LOGGER.debug("Event received: %s", resp.state_update)

    @property
    def state(self) -> str:
        """Get the current state of the connection."""
        return self._state


class ExpectedResponse:
    """Define an awaitable command event response."""

    def __init__(self, command: str, val: str):
        """Init a new instance of the CommandEvent."""
        self._event = asyncio.Event()
        self._command = command
        self._val = val
        self._response = None
        self._time_sent = datetime.utcnow()

    async def wait(self):
        """Wait until the event is set."""
        await self._event.wait()
        return self._response

    def set(self, message: Message):
        """Set the response."""
        self._response = message
        self._event.set()

    @property
    def command(self) -> int:
        """Get the command that represents this event."""
        return self._command



"""Define commands."""
from abc import ABC, abstractmethod

from pyavreceiver.error import AVReceiverInvalidArgumentError


def identity(arg, **kwargs):
    """The identity function returns the input."""
    # pylint: disable=unused-argument
    return arg


class CommandValues:
    """Possible values for a command."""

    def __init__(self, values: dict):
        self._values = values
        self._values["min"] = self._values.get("min")
        self._values["max"] = self._values.get("max")

    def __repr__(self):
        return str(self._values)

    def __str__(self):
        return str({x for x in self._values if self._values[x] is not None})

    def get(self, name):
        """Patch to dict.get()."""
        return self._values.get(name)

    def update(self, _dict):
        """Patch to dict.update()."""
        self._values.update(_dict)

    def __getattr__(self, name: str) -> str:
        if name in self._values:
            return self._values[name]
        raise AVReceiverInvalidArgumentError

    def __getitem__(self, name: str) -> str:
        if name in self._values:
            return self._values[name]
        raise AVReceiverInvalidArgumentError

    def __setitem__(self, name: str, val):
        """Only set if name does not exist."""
        if name not in self._values:
            self._values[name] = val


class TelnetCommand(ABC):
    """Define the telnet command interface."""

    def __init__(
        self,
        *,
        name: str,
        command: str,
        values: CommandValues,
        val_pfx: str = "",
        func=identity,
        zero: int = 0,
        val=None,
        valid_strings: list = None,
        message: str = None,
    ):
        self._name = name
        self._command = command
        self._values = values
        self._val_pfx = val_pfx
        self._func = func
        self._zero = zero
        self._val = val
        self._valid_strings = valid_strings

        # self._val_translate = {"True": "ON", "False": "OFF"}
        self._message = message

    @abstractmethod
    def set_val(self, val) -> str:
        """Format the command with argument and return."""

    @abstractmethod
    def set_query(self) -> str:
        """Format the command with query and return."""

    @property
    @abstractmethod
    def command(self) -> str:
        """The command portion of the message."""

    @property
    @abstractmethod
    def message(self) -> str:
        """The complete message; command + argument."""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the command."""

    @property
    @abstractmethod
    def val(self) -> str:
        """The argument of the command."""

    @property
    @abstractmethod
    def values(self) -> list:
        """Return the valid argument values."""



"""Define constants."""

__title__ = "pyavreceiver"
__version__ = "0.0.2"


CLI_PORT = 23
DEFAULT_TIMEOUT = 10.0
DEFAULT_RECONNECT_DELAY = 10.0
DEFAULT_HEART_BEAT = 11.0
DEFAULT_STEP = 5

STATE_CONNECTED = "connected"
STATE_DISCONNECTED = "disconnected"
STATE_RECONNECTING = "reconnecting"

# Using ^ as escape character
COMMAND_NAME = "^name"
COMMAND_NAMES = "^names"
COMMAND_PARAMS = "^params"
COMMAND_RANGE = "^range"
COMMAND_FUNCTION = "^function"
COMMAND_STRINGS = "^strings"
COMMAND_ZERO = "^zero"

DEVICE_INFO_ENDPOINTS = [
    ":80/goform/Deviceinfo.xml",  # Denon/Marantz
    ":8080/goform/Deviceinfo.xml",  # Denon/Marantz
    "/YamahaRemoteControl/ctrl?",  # Yamaha
]

EVENT_TELNET = "telnet"
EVENT_DISCONNECTED = "disconnected"
EVENT_CONNECTED = "connected"

SIGNAL_STATE_UPDATE = "state_update"
SIGNAL_TELNET_EVENT = "telnet_event"
SIGNAL_STATE_UPDATE = "state_update"


ATTR_POWER = "power"
ATTR_VOLUME = "volume"
ATTR_VOLUME_UP = "volume_up"
ATTR_VOLUME_DOWN = "volume_down"
ATTR_LFE_LEVEL = "lfe_level"
ATTR_MUTE = "mute"
ATTR_SOURCE = "source"
ATTR_SOUND_MODE = "sound_mode"
ATTR_SUBWOOFER_ONE = "subwoofer_1"
ATTR_TONE_CONTROL = "tone_control"
ATTR_TREBLE = "treble"
ATTR_BASS = "bass"
ATTR_DIALOG_LEVEL = "dialog_level"
ATTR_DSP_DRC = "dsp_dynamic_range_control"
ATTR_DSP_MODE = "dsp_mode"
ATTR_META_DRC = "meta_dynamic_range_control"
ATTR_ZONE1_POWER = "zone1_power"

ATTR_ZONE2_POWER = "zone2_power"
ATTR_ZONE2_VOLUME = "zone2_volume"
ATTR_ZONE2_VOLUME_UP = "zone2_volume_up"
ATTR_ZONE2_VOLUME_DOWN = "zone2_volume_down"
ATTR_ZONE2_BASS = "zone2_bass"
ATTR_ZONE2_TREBLE = "zone2_treble"
ATTR_ZONE2_SOURCE = "zone2_source"

ATTR_ZONE3_POWER = "zone3_power"
ATTR_ZONE3_VOLUME = "zone3_volume"
ATTR_ZONE3_VOLUME_UP = "zone3_volume_up"
ATTR_ZONE3_VOLUME_DOWN = "zone3_volume_down"
ATTR_ZONE3_BASS = "zone3_bass"
ATTR_ZONE3_TREBLE = "zone3_treble"
ATTR_ZONE3_SOURCE = "zone3_source"

VAL_DOWN = "down"
VAL_UP = "up"

ZONE_PREFIX = {
    "main": "",
    "zone1": "",
    "zone_1": "",
    "zone2": "zone2_",
    "zone_2": "zone2_",
    "zone3": "zone3_",
    "zone_3": "zone3_",
    "zone4": "zone4_",
    "zone_4": "zone4_",
}

FUNCTION_VOLUME = "volume"
FUNCTION_NUM_TO_DB = "num_to_db"
FUNCTION_DB_TO_NUM = "db_to_num"

INFO_MODEL = "model_name"
INFO_MAC = "mac_address"
INFO_ZONES = "zones"
INFO_SERIAL = "serial_number"

MESSAGE_INTERVAL_LIMIT = 50  # milliseconds

UPNP_ENDPOINTS = {
    "denon-avr": ":8080/description.xml",
    "denon-avr-x": ":8080/description.xml",
    "denon-avr-x-2016": ":60006/upnp/desc/aios_device/aios_device.xml",
}



"""Define helpers for parsing integer values returned by receiver."""
import math

from pyavreceiver.denon.error import DenonCannotParse


class Parse:
    """Define parsers for string integer to db and back."""

    def __getitem__(self, key):
        return getattr(self, key)

    @staticmethod
    def num_to_db(num: str = None, zero: int = 80, valid_strings=None):
        """Convert the string num to a decibel float."""
        zero = 0 if zero is None else zero
        if valid_strings and num in valid_strings:
            return valid_strings[num]
        try:
            if len(num) == 3 and num[0] != "-":
                vol = int(num) / 10
            elif int(num) < 100 and int(num) > -100:
                vol = int(num)
            else:
                raise DenonCannotParse

            return vol - zero
        except TypeError:
            return num
        except ValueError:
            return None

    @staticmethod
    def db_to_num(
        decibel: int = None, zero: int = 80, str_len: int = 0, valid_strings=None
    ):
        """Convert the float decibel to a string num."""
        try:
            zero = 0 if zero is None else zero

            decibel = round(decibel * 2)
            if decibel != 0:
                decibel /= 2

            vol = decibel + zero
            if vol - math.ceil(vol) != 0:
                vol *= 10

            out = str(int(vol))

            if padding := str_len - len(out) > 0:
                pad = ""
                while padding > -1:
                    pad += "0"
                    padding -= 1
                return pad + out
            return out
        except TypeError:
            return decibel


parse = Parse()



"""Define Denon/Marantz commands."""
from collections import defaultdict

import pyavreceiver.denon.const as denon_const
from pyavreceiver import const
from pyavreceiver.command import CommandValues, TelnetCommand, identity
from pyavreceiver.denon.parse import parse


class DenonTelnetCommand(TelnetCommand):
    """Representation of a Denon telnet message command."""

    def set_val(self, val=None) -> str:
        """Format the command with argument and return."""
        if val is not None:
            val = (
                self._values.get(bool(val))
                or self._values.get(val)
                or self._values.get(str(val).lower())
                or val
            )
            val = "ON" if val is True else val
            val = "OFF" if val is False else val
            try:
                val = val.upper()
            except AttributeError:
                pass
            message = (
                f"{self._command}{self._val_pfx}"
                f"{self._func(val, zero=self._zero, valid_strings=self._valid_strings)}"
                f"{denon_const.TELNET_SEPARATOR}"
            )
        else:
            message = f"{self._command}{denon_const.TELNET_SEPARATOR}"
        return DenonTelnetCommand(
            name=self._name,
            command=self._command,
            values=self._values,
            val_pfx=self._val_pfx,
            func=self._func,
            zero=self._zero,
            val=val,
            message=message,
            valid_strings=self._valid_strings,
        )

    def set_query(self) -> str:
        """Format the command with query and return."""
        message = (
            f"{self._command}{self._val_pfx}{denon_const.TELNET_QUERY}"
            f"{denon_const.TELNET_SEPARATOR}"
        )
        return DenonTelnetCommand(
            name=self._name,
            command=self._command,
            values=self._values,
            val_pfx=self._val_pfx,
            func=self._func,
            zero=self._zero,
            val=denon_const.TELNET_QUERY,
            message=message,
            valid_strings=self._valid_strings,
        )

    @property
    def command(self) -> str:
        """The command portion of the message."""
        return self._command

    @property
    def message(self) -> str:
        """The complete message; command + argument."""
        if not self._message:
            raise Exception
        return self._message

    @property
    def name(self) -> str:
        """The name of the command."""
        return self._name

    @property
    def val(self) -> str:
        """The argument of the command."""
        return self._val

    @property
    def values(self) -> list:
        """Return the valid argument values."""
        return self._values


def get_command_lookup(command_dict):
    """Return the command lookup dict."""
    command_lookup = defaultdict(None)
    for cmd, entry in command_dict.items():
        try:
            val_range = entry.get(const.COMMAND_RANGE)
            zero = entry.get(const.COMMAND_ZERO)
            func = entry.get(const.COMMAND_FUNCTION)
            func = parse["num_to_db"] if func == const.FUNCTION_VOLUME else None
            valid_strings = entry.get(const.COMMAND_STRINGS)
        except AttributeError:
            val_range, zero, func, valid_strings = None, None, None, None

        if const.COMMAND_PARAMS not in entry:
            # No nested params, make entry and continue loop
            val_pfx = ""
            if const.COMMAND_NAME in entry:
                names = [entry[const.COMMAND_NAME]]
            elif const.COMMAND_NAMES in entry:
                names = entry[const.COMMAND_NAMES].values()
            else:
                names = [cmd]
            for name in names:
                add_command(
                    command_lookup,
                    entry,
                    name,
                    cmd,
                    val_pfx,
                    val_range,
                    zero,
                    func,
                    valid_strings,
                )
            continue

        if const.COMMAND_PARAMS in entry and const.COMMAND_NAME in entry:
            # Nested params and main entry, make entry and stay in loop
            val_pfx = ""
            if const.COMMAND_NAME in entry:
                name = entry[const.COMMAND_NAME]
            else:
                name = cmd
            add_command(
                command_lookup,
                entry,
                name,
                cmd,
                val_pfx,
                val_range,
                zero,
                func,
                valid_strings,
            )

        # Nested params, iterate and make entries
        val_pfx = " "
        for prm, sub_entry in entry.items():
            if prm.startswith("^"):
                # prm is not a command
                continue
            try:
                sub_val_range = sub_entry.get(const.COMMAND_RANGE)
                sub_range_zero = sub_entry.get(const.COMMAND_ZERO)
                sub_func = sub_entry.get(const.COMMAND_FUNCTION)
                sub_func = (
                    parse["num_to_db"] if sub_func == const.FUNCTION_VOLUME else func
                )
                sub_valid_strings = sub_entry.get(const.COMMAND_STRINGS)
            except AttributeError:
                sub_val_range, sub_range_zero, sub_func, sub_valid_strings = (
                    None,
                    None,
                    None,
                    None,
                )
            if sub_entry and const.COMMAND_NAME in sub_entry:
                full_name = sub_entry[const.COMMAND_NAME]
            elif name := entry.get(const.COMMAND_NAME):
                full_name = f"{name}_{prm.lower()}"
            else:
                full_name = f"{cmd}_{prm}"
            command = f"{cmd}{prm}"
            add_command(
                command_lookup,
                sub_entry,
                full_name,
                command,
                val_pfx,
                sub_val_range or val_range,
                sub_range_zero or zero,
                sub_func or func,
                sub_valid_strings or valid_strings,
            )
    return command_lookup


def add_command(ref, entry, name, cmd, val_pfx, val_range, zero, func, valid_strings):
    """Add the command to the command dictionary, ref."""
    values = {}
    try:
        for item in entry:
            if str(item).startswith("^"):
                continue
            if entry[item] is not None:
                try:
                    values[entry[item].lower()] = item
                    if entry[item].lower() == "off":
                        values[False] = item
                    elif entry[item].lower() == "on":
                        values[True] = item
                except AttributeError:
                    values[entry[item]] = item
            values[item.lower()] = item
    except (TypeError, AttributeError):
        pass
    func = func or identity
    if val_range is not None:
        val_range = {
            "min": func(str(val_range[0]), zero=zero),
            "max": func(str(val_range[1]), zero=zero),
        }
    if values and val_range:
        values = values.update(val_range)
    values = val_range or values
    ref[name] = DenonTelnetCommand(
        name=name,
        command=cmd,
        values=CommandValues(values),
        val_pfx=val_pfx,
        func=parse["db_to_num"] if func else None,
        zero=zero,
        valid_strings=valid_strings,
    )



"""pyavreceiver errors."""


class AVReceiverInvalidArgumentError(Exception):
    """Invalid argument error."""


class AVReceiverIncompatibleDeviceError(Exception):
    """Invalid argument error."""



"""Implement a Denon telnet message."""
import logging

from pyavreceiver import const
from pyavreceiver.denon.error import DenonCannotParse
from pyavreceiver.denon.parse import parse
from pyavreceiver.response import Message

_LOGGER = logging.getLogger(__name__)


class DenonMessage(Message):
    """Define a Denon telnet message representation."""

    def __init__(self, message: str = None, command_dict: dict = None):
        """Init a new Denon message."""
        self._message = None  # type: str
        self._raw_val = None  # type: str
        self._cmd = None  # type: str
        self._prm = None  # type: str
        self._val = None
        self._name = None  # type: str
        self._command_dict = command_dict or {}
        self._new_command = None

        self._state_update = self._parse(message) if message else {}

    def __str__(self):
        """Get user readable message."""
        return self._message

    def __repr__(self):
        """Get readable message."""
        return self._message

    def _parse(self, message: str) -> dict:
        """Parse message, assign attributes, return a state update dict."""
        self._message = message
        self._cmd, self._prm, self._raw_val = self.separate(message)
        try:
            self._val = self.parse_value(self._cmd, self._prm, self._raw_val)
        except DenonCannotParse:
            return {}
        return self._make_state_update()

    def _make_state_update(self) -> dict:
        """Return the state update dict."""
        entry = self._command_dict.get(self._cmd) or self._cmd
        key = entry
        val = self._val if self._val is not None else self._raw_val
        try:
            if const.COMMAND_NAMES in entry:
                try:
                    _ = int(val)
                    val_type = "number"
                except ValueError:
                    val_type = val
                key = entry[const.COMMAND_NAMES].get(val_type)
                if key is None:
                    key = entry[const.COMMAND_NAMES]["other"]

            if const.COMMAND_NAME in entry and const.COMMAND_PARAMS in entry:
                cmd_name = entry[const.COMMAND_NAME]
                key = f"{cmd_name}_{self._prm.lower()}" if self._prm else cmd_name
                entry = entry.get(self._prm)

            try:
                key = (
                    entry.get(const.COMMAND_NAME)
                    or key.get(const.COMMAND_NAME)  # key may be string
                    or f"{self._cmd}{'_' + self._prm if self._prm else ''}"
                )
            except AttributeError:
                pass

            _ = entry.get(self._raw_val)
            val = _ if _ is not None else self._val or val
            entry = entry.get(self._prm)
            val = entry.get(self._raw_val) or self._val or val

            key = (
                self._command_dict[self._cmd][self._prm].get(const.COMMAND_NAME) or key
            )

        except (KeyError, AttributeError):
            pass
        self._name = key
        return {key: val}

    def separate(self, message) -> tuple:
        """Separate command category, parameter, and value."""
        return DenonMessage._search(self._command_dict, 0, message, (), self=self)

    @staticmethod
    def _search(lvl: dict, depth: int, rem: str, cur: tuple, self=None) -> tuple:
        """Search dict for best match."""
        # pylint: disable=protected-access
        if rem in lvl:
            if depth == 1:
                return (*cur, None, rem)
        elif const.COMMAND_RANGE in lvl and rem.isnumeric():
            return (*cur, None, rem)
        elif const.COMMAND_PARAMS in lvl:
            prm = rem
            val = ""
            # Check for match at each partition
            for _ in range(len(prm)):
                if prm in lvl:
                    return (*cur, prm.strip(), val.strip())
                val = prm[-1:] + val
                prm = prm[:-1]
            # No match found: return new entry, assume val after last space
            words = rem.split(" ")
            if len(words) < 2:
                _LOGGER.info(
                    "Added new event with empty value: %s, %s, None", *cur, rem
                )
                if self:
                    self._new_command = {"cmd": cur[0], "prm": rem, "val": None}
                return (*cur, words[0], None)
            prm = " ".join(words[:-1]).strip()
            val = words[-1].strip()
            _LOGGER.info("Added new event: %s, %s, %s", *cur, prm, val)
            if self:
                self._new_command = {"cmd": cur[0], "prm": prm, "val": val}
            return (*cur, prm, val)
        elif depth == 1:
            self._new_command = {"cmd": cur[0], "prm": None, "val": rem.strip()}
            return (*cur, None, rem.strip())
        # Search for matches at every prefix/postfix
        for i in range(-1, -len(rem), -1):
            prefix = rem[:i]
            if prefix not in lvl:
                continue
            return DenonMessage._search(
                lvl[prefix], depth + 1, rem[i:], (*cur, prefix), self=self
            )
        # No match found: return new entry, assume val after last space
        words = rem.split(" ")
        if len(words) < 2:
            _LOGGER.error("Unparsable event: %s", rem)
            return (rem, None, None)
        cmd = " ".join(words[:-1]).strip()
        val = words[-1].strip()
        _LOGGER.info("Parsed new cmd event: %s, None, %s", cmd, val)
        if self:
            self._new_command = {"cmd": cmd, "prm": None, "val": val}
        return (cmd, None, val)

    def parse_value(self, cmd: str, prm: str, val: str):
        """Parse a value from val."""
        if not isinstance(self._command_dict.get(cmd), dict):
            return val
        entry = self._command_dict[cmd].get(prm) or self._command_dict[cmd]
        if const.COMMAND_FUNCTION in entry:
            function_name = entry[const.COMMAND_FUNCTION]
        elif const.COMMAND_FUNCTION in self._command_dict[cmd]:
            function_name = self._command_dict[cmd][const.COMMAND_FUNCTION]
        else:
            return val
        if function_name == const.FUNCTION_VOLUME:
            parser = parse[const.FUNCTION_NUM_TO_DB]
            return parser(
                num=val,
                zero=entry.get(const.COMMAND_ZERO),
                valid_strings=entry.get(const.COMMAND_STRINGS),
            )
        raise Exception

    @property
    def parsed(self) -> tuple:
        """Return the message parsed into (command, param, value)."""
        return (self._cmd, self._prm, self._val)

    @property
    def message(self) -> str:
        return self._message

    @property
    def raw_value(self) -> str:
        return self._raw_val

    @property
    def state_update(self) -> dict:
        return self._state_update

    @property
    def command(self) -> str:
        return self._cmd + (self._prm or "")

    @property
    def new_command(self) -> dict:
        return self._new_command

    @property
    def name(self) -> str:
        return self._name



"""pyavreceiver"""

import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join("README.md"), "r") as fh:
    long_description = fh.read()

const = {}
with open(os.path.join("pyavreceiver", "const.py"), "r") as fp:
    exec(fp.read(), const)

setup(
    name=const["__title__"],
    version=const["__version__"],
    description="An async python library for controlling Audio Video Receiver devices",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jphutchins/pyavreceiver",
    author="J.P. Hutchins",
    author_email="jphutchins@gmail.com",
    license="ASL 2.0",
    packages=find_packages(exclude=("tests", "tests.*")),
    install_requires=[
        "aiofiles",
        "aiohttp",
        "PyYAML",  
    ],
    tests_require=["tox>=3.5.0,<4.0.0"],
    platforms=["any"],
    keywords="denon marantz receiver",
    zip_safe=False,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
        "Topic :: Home Automation",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires='>=3.8',
)



"""Define an HTTP connection to a Denon/Marantz receiver."""
from xml.etree import ElementTree as ET

import aiohttp

from pyavreceiver import const
from pyavreceiver.denon import const as denon_const
from pyavreceiver.http_connection import HTTPConnection


class DenonHTTPApi(HTTPConnection):
    """Define the Denon HTTP API."""

    async def get_device_info(self) -> dict:
        "Get information about the device"
        xml = await self._get_device_info()
        self.make_device_info_dict(xml)
        if self._upnp_data:
            self.make_device_info_dict(self._upnp_data)
        return self.device_info

    async def get_source_names(self) -> dict:
        """Get the renamed sources."""
        xml_body = DenonHTTPApi.make_xml_request(
            ["GetRenameSource", "GetDeletedSource"]
        )
        xml = await self._app_command(xml_body)
        if xml:
            return DenonHTTPApi.make_renamed_dict(xml)

        # AppCommand endpoint failed, use alternate
        xml = await self._get_mainzone_xml() or await self._get_status_xml()
        return DenonHTTPApi.make_renamed_dict_legacy(xml)

    async def _get_status_xml(self) -> str:
        """Get the Main Zone status XML endpoint."""
        async with aiohttp.ClientSession() as session:
            async with session.get(
                f"http://{self.host}:{self.port}{denon_const.API_MAIN_ZONE_XML_STATUS_URL}"
            ) as resp:
                if resp.status == 200:
                    return await resp.text()

    async def _get_mainzone_xml(self) -> str:
        """Get the Main Zone status XML endpoint."""
        async with aiohttp.ClientSession() as session:
            async with session.get(
                f"http://{self.host}:{self.port}{denon_const.API_MAIN_ZONE_XML_URL}"
            ) as resp:
                if resp.status == 200:
                    return await resp.text()

    async def _get_device_info(self):
        """Get information about the device."""
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"http://{self.host}:{self.port}{self._device_info_url}"
            ) as resp:
                if resp.status == 200:
                    return await resp.text()

    async def _app_command(self, xml: bytes):
        """Make request to AppCommand.xml endpoint."""
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"http://{self.host}:{self.port}/goform/AppCommand.xml", data=xml
            ) as resp:
                if resp.status == 200:
                    return await resp.text()
                return False

    @staticmethod
    def make_renamed_dict(xml) -> dict:
        """Parse the XML response for renamed and deleted sources."""
        root = ET.fromstring(xml)
        deleted = set()
        deleted_root = root.find("*/functiondelete")
        for entry in deleted_root.findall("list"):
            try:
                if entry.find("use").text == "0":
                    deleted.add(entry.find("name").text.strip().lower())
            except AttributeError:
                continue
        rename_map = {}
        renamed_root = root.find("*/functionrename")
        for entry in renamed_root.findall("list"):
            try:
                name = entry.find("name").text.strip().lower()
                rename = entry.find("rename").text.strip()
            except AttributeError:
                continue
            if name not in deleted:
                rename_map[rename] = (
                    denon_const.MAP_HTTP_SOURCE_NAME_TO_TELNET.get(name) or name.upper()
                )
        return rename_map

    @staticmethod
    def make_renamed_dict_legacy(xml) -> dict:
        """Parse the XML response for renamed and deleted sources."""
        root = ET.fromstring(xml)

        original_names = []
        skip_source = 0
        for name in root.find("InputFuncList"):
            try:
                if name.text == "SOURCE":
                    skip_source = 1
                    continue
                original_names.append(name.text)
            except AttributeError:
                continue

        deleted = [False] * len(original_names)

        delete_list = root.find("SourceDelete")
        if delete_list:
            for i, delete in enumerate(delete_list[skip_source:]):
                try:
                    if delete.text == "DEL":
                        deleted[i] = True
                except AttributeError:
                    continue

        rename_map = {}
        for i, name in enumerate(root.find("RenameSource")[skip_source:]):
            if not deleted[i]:
                original_name = original_names[i]
                try:
                    name = name.text.strip()
                except AttributeError:
                    name = original_name
                rename_map[name] = (
                    denon_const.MAP_HTTP_SOURCE_NAME_TO_TELNET.get(
                        original_name.lower()
                    )
                    or original_name.upper()
                )
        return rename_map

    def make_device_info_dict(self, xml) -> dict:
        """Parse response for information."""
        root = ET.fromstring(xml)
        xmlns = "{urn:schemas-upnp-org:device-1-0}"
        self._device_info[const.INFO_MODEL] = (
            self._device_info.get(const.INFO_MODEL)
            or get_text(root, "ModelName")
            or get_text(root, f"*/{xmlns}modelName")
        )
        self._device_info[const.INFO_MAC] = get_text(root, "MacAddress")
        self._device_info[const.INFO_SERIAL] = get_text(root, f"*/{xmlns}serialNumber")
        self._device_info[const.INFO_ZONES] = get_text(root, "DeviceZones") or "0"

    @staticmethod
    def make_xml_request(commands: list) -> bytes:
        """Prepare XML body for Denon API."""
        xml_parts = ['<?xml version="1.0" encoding="utf-8"?>\n', "<tx>"]
        cmd_l = '<cmd id="1">'
        cmd_r = "</cmd>"
        for i, command in enumerate(commands):
            if i != 0 and i % 5 == 0:
                # API allows multiple XML roots, limit 5 commands each
                xml_parts.append("</tx><tx>")
            xml_parts.append(f"{cmd_l}{command}{cmd_r}")
        xml_parts.append("</tx>")
        return "".join(xml_parts).encode()


class DenonAVRApi(DenonHTTPApi):
    """Define the Denon/Marantz AVR API."""

    def __init__(self, host, upnp_data):
        super().__init__(host, upnp_data=upnp_data)
        self.port = denon_const.API_PORT
        self._device_info_url = denon_const.API_DEVICE_INFO_URL


class DenonAVRXApi(DenonHTTPApi):
    """Define the Denon/Marantz AVR-X API."""

    def __init__(self, host, upnp_data):
        super().__init__(host, upnp_data=upnp_data)
        self.port = denon_const.API_PORT
        self._device_info_url = denon_const.API_DEVICE_INFO_URL


class DenonAVRX2016Api(DenonHTTPApi):
    """Define the Denon/Marantz AVR-X 2016 API."""

    def __init__(self, host, upnp_data):
        super().__init__(host, upnp_data=upnp_data)
        self.port = denon_const.API_2016_PORT
        self._device_info_url = denon_const.API_2016_DEVICE_INFO_URL


def get_text(root: ET.Element, tag: str):
    """Get the text attribute if it exists."""
    elem = root.find(tag)
    if elem is not None:
        return elem.text
    return None



"""pyavreceiver - interface to control Audio/Video Receivers."""
import asyncio

import aiohttp

from pyavreceiver import const
from pyavreceiver.denon.http_connection import (
    DenonAVRApi,
    DenonAVRX2016Api,
    DenonAVRXApi,
)
from pyavreceiver.denon.receiver import DenonReceiver
from pyavreceiver.error import AVReceiverIncompatibleDeviceError


async def factory(host: str):
    """Return an instance of an AV Receiver."""

    tasks = []
    names = []
    async with aiohttp.ClientSession() as session:
        for name, url in const.UPNP_ENDPOINTS.items():
            names.append(name)
            tasks.append(
                asyncio.create_task(session.get(f"http://{host}{url}", timeout=1))
            )
        results = await asyncio.gather(*tasks)
    for name, response in zip(names, results):
        if response.status == 200:
            if name == "denon-avr-x-2016":
                http_api = DenonAVRX2016Api(host, await response.text())
                return DenonReceiver(host, http_api=http_api)
            if name == "denon-avr-x":
                http_api = DenonAVRXApi(host, await response.text())
                return DenonReceiver(host, http_api=http_api)
            if name == "denon-avr":
                http_api = DenonAVRApi(host, await response.text())
                return DenonReceiver(host, http_api=http_api)
        raise AVReceiverIncompatibleDeviceError



"""Define the interface of an A/V Receiver Zone."""
from pyavreceiver import const


class Zone:
    """Define an Audio/Video Receiver zone."""

    def __init__(self, avr, zone: str = "main"):
        """Init the zone."""
        self._avr = avr
        self._zone_prefix = const.ZONE_PREFIX[zone.lower()]
        self._commands = avr._connection._command_lookup  # type: dict

    def get(self, name: str) -> str:
        """Get the current state of the nameibute name."""
        return self._avr.state.get(self._zone_prefix + name)

    def set(self, name: str, val=None) -> bool:
        """Request the receiver set the name to val."""
        # pylint: disable=protected-access
        if not self._avr.power:
            return False
        command = self._avr._connection._command_lookup[
            f"{self._zone_prefix}{name}"
        ].set_val(val)
        self._avr._connection.send_command(command)
        return True

    def update(self, name: str):
        """Request the receiver to send update of the value of name."""
        # pylint: disable=protected-access
        command = self._avr._connection._command_lookup[name].set_query()
        self._avr._connection.send_command(command)

    def update_all(self):
        """Update all known attributes in commands."""
        for name in self.commands:
            self.update(name)

    @property
    def commands(self) -> dict:
        """The dictionary of available commands."""
        return self._commands

    @property
    def bass(self) -> int:
        """The state of bass."""
        return self.get(const.ATTR_BASS)

    def set_bass(self, val: float) -> bool:
        """Request the receiver set the bass to val."""
        return self.set(const.ATTR_BASS, val)

    @property
    def mute(self) -> str:
        """The state of mute."""
        return self.get(const.ATTR_MUTE)

    def set_mute(self, val: bool) -> bool:
        """Request the receiver set mute to val."""
        return self.set(const.ATTR_MUTE, val)

    @property
    def power(self) -> str:
        """The state of zone power."""
        return self.get(const.ATTR_ZONE1_POWER)

    def set_power(self, val: bool) -> bool:
        """Request the receiver set zone power to val."""
        return self.set(const.ATTR_ZONE1_POWER, val)

    @property
    def source(self) -> str:
        """The state of source."""
        return self.get(const.ATTR_SOURCE)

    def set_source(self, val: str) -> bool:
        """Request the receiver set the source to val."""
        return self.set(const.ATTR_SOURCE, val)

    @property
    def treble(self) -> int:
        """The state of treble."""
        return self.get(const.ATTR_TREBLE)

    def set_treble(self, val: float) -> bool:
        """Request the receiver set the treble to val."""
        return self.set(const.ATTR_TREBLE, val)

    @property
    def volume(self) -> str:
        """The state of volume."""
        return self.get(const.ATTR_VOLUME)

    def set_volume(self, val: float) -> bool:
        """Request the receiver set volume to val."""
        return self.set(const.ATTR_VOLUME, val)

    def set_volume_down(self) -> bool:
        """Request the receiver turn the volume down."""
        return self.set(const.ATTR_VOLUME_DOWN)

    def set_volume_up(self) -> bool:
        """Request the receiver turn the volume up."""
        return self.set(const.ATTR_VOLUME_UP)


class MainZone(Zone):
    """Define an Audio/Video Receiver main zone.

    Attributes are ordered alphabetically.  Setters are prefixed with set_ and grouped with
    their corresponding property.
    """

    @property
    def dialog_level(self) -> str:
        """The state of dialog level."""
        return self.get(const.ATTR_DIALOG_LEVEL)

    def set_dialog_level(self, val) -> bool:
        """Request the receiver set dialog level to val."""
        return self.set(const.ATTR_DIALOG_LEVEL, val)

    @property
    def dsp_dynamic_range(self) -> str:
        """The state of DSP dynamic range."""
        return self.get(const.ATTR_DSP_DRC)

    def set_dsp_dynamic_range(self, val) -> bool:
        """Request the receiver set the DRC to val."""
        return self.set(const.ATTR_DSP_DRC, val)

    @property
    def dsp_mode(self) -> str:
        """The state of DSP mode."""
        return self.get(const.ATTR_DSP_MODE)

    def set_dsp_mode(self, val: str) -> bool:
        """Request the receiver set DSP mode to val."""
        return self.set(const.ATTR_DSP_MODE, val)

    @property
    def lfe_level(self) -> str:
        """The state of LFE level."""
        return self.get(const.ATTR_LFE_LEVEL)

    def set_lfe_level(self, val: float) -> bool:
        """Request the receiver set LFE level to val."""
        return self.set(const.ATTR_LFE_LEVEL, val)

    @property
    def meta_dynamic_range(self) -> str:
        """The state of metadata dynamic range control."""
        return self.get(const.ATTR_META_DRC)

    def set_meta_dynamic_range(self, val: str) -> bool:
        """Request the receiver set meta DRC to val."""
        return self.set(const.ATTR_META_DRC, val)

    @property
    def soundmode(self) -> str:
        """The state of soundmode."""
        return self.get(const.ATTR_SOUND_MODE)

    def set_soundmode(self, val: str) -> bool:
        """Request the receiver set the sound mode to val."""
        return self.set(const.ATTR_SOUND_MODE, val)

    @property
    def subwoofer_one(self) -> bool:
        """The state of subwoofer one."""
        return self.get(const.ATTR_SUBWOOFER_ONE)

    def set_subwoofer_one(self, val: bool) -> bool:
        """Request the receiver set subwoofer one to val."""
        return self.set(const.ATTR_SUBWOOFER_ONE, val)

    @property
    def tone_control(self) -> str:
        """The state of tone control."""
        return self.get(const.ATTR_TONE_CONTROL)

    def set_tone_control(self, val: bool) -> bool:
        """Request the receiver set the treble to val."""
        return self.set(const.ATTR_TONE_CONTROL, val)



"""Define constants for Denon/Marantz."""

API_DEVICE_INFO_URL = "/goform/Deviceinfo.xml"
API_MAIN_ZONE_XML_STATUS_URL = "/goform/formMainZone_MainZoneXmlStatus.xml"
API_MAIN_ZONE_XML_URL = "/goform/formMainZone_MainZoneXml.xml"
API_PORT = 80

API_2016_PORT = 8080
API_2016_DEVICE_INFO_URL = "/goform/Deviceinfo.xml"

CLI_PORT = 23

MESSAGE_INTERVAL_LIMIT = 50  # milliseconds
MESSAGE_RESPONSE_TIMEOUT = 250  # milliseconds

TELNET_SEPARATOR = "\r"

TELNET_PORT = 23
DEFAULT_TIMEOUT = 10.0
DEFAULT_RECONNECT_DELAY = 10.0
DEFAULT_HEART_BEAT = 10.0

DEVICE_INFO_ENDPOINTS = [
    ":80/goform/Deviceinfo.xml",
    ":8080/goform/Deviceinfo.xml",
]

DEVICE_INFO_SEARCH = {"model": ["ModelName", "ManualModelName"]}

SOURCE_AUX = "AUX1"
SOURCE_BLURAY = "BD"
SOURCE_BLUETOOTH = "BT"
SOURCE_CABLE = "SAT/CBL"
SOURCE_DVD = "DVD"
SOURCE_IPOD = "IPD"
SOURCE_USB_IPOD = "USB/IPOD"
SOURCE_GAME = "GAME"
SOURCE_MEDIA_PLAYER = "MPLAY"
SOURCE_NETWORK = "NET"
SOURCE_PHONO = "PHONO"
SOURCE_TUNER = "TUNER"
SOURCE_TV_AUDIO = "TV"
SOURCE_USB = "USB"

MAP_HTTP_SOURCE_NAME_TO_TELNET = {
    "aux": SOURCE_AUX,
    "blu-ray": SOURCE_BLURAY,
    "bluetooth": SOURCE_BLUETOOTH,
    "cbl/sat": SOURCE_CABLE,
    "dvd": SOURCE_DVD,
    "fm": SOURCE_TUNER,
    "ipod": SOURCE_IPOD,
    "ipod/usb": SOURCE_USB_IPOD,
    "game": SOURCE_GAME,
    "media player": SOURCE_MEDIA_PLAYER,
    "network": SOURCE_NETWORK,
    "phono": SOURCE_PHONO,
    "tuner": SOURCE_TUNER,
    "tv audio": SOURCE_TV_AUDIO,
    "usb": SOURCE_USB,
}

TELNET_QUERY = "?"

ATTR_DYNAMIC_EQ = "audyssey_dynamic_eq"

XML_MODEL_NAME = "ModelName"
XML_MAC_ADDRESS = "MacAddress"
XML_ZONE_COUNT = "DeviceZones"
XML_API_VERS = "CommApiVers"
